// Copyright The pipewire-rs Contributors.
// SPDX-License-Identifier: MIT

//! Pipewire constants.

/// Invalid ID that matches any object when used for permissions.
pub const ID_ANY: u32 = 0xffffffff;
